@@ -1,5 +1,6 @@
 import sys
 from sly import Lexer, Parser
+from .logger import print_info, print_warn, print_error, info as _info, warn as _warn, error as _error
 
 print(end="zink: ... ", flush=True)
 
@@ -7,49 +8,63 @@ class FilteredLogger(object):
     def __init__(self, f, exit: bool = True):
         self.f = f
         self.exit = exit
+        self.had_warnings = False
+        self.had_errors = False
 
     def debug(self, msg, *args, **kwargs):
-        self.f.write((msg % args) + '\n')
+        self.f.write(_info((msg % args) + '\n'))
 
     info = debug
 
     def warning(self, msg, *args, **kwargs):
+        self.had_warnings = True
         tolog = (msg % args)
         if "defined, but not used" in tolog: return
         if "unused tokens" in tolog: return
-        self.f.write("WARNING: " + tolog + "\n")
+        self.f.write(_warn(tolog + "\n"))
         if self.exit and "conflict" in tolog: exit(1)
 
     def error(self, msg, *args, **kwargs):
-        self.f.write("ERROR: " + (msg % args) + "\n")
+        self.had_errors = True
+        self.f.write(_error((msg % args) + "\n"))
 
-    critical = debug
+    critical = error
+
+    def reset_warnings(self):
+        self.had_warnings = False
+
+    def reset_errors(self):
+        self.had_errors = False
 
 class ZinkLexer(Lexer):
     tokens = {
         "ID", "NUMBER", "STRING", "BSTRING", "RSTRING", "RAWSTRING", "TRUE", "FALSE", "NONE",
         "EQUAL",
         "DB_PLUS", "DB_MINUS",
-        "PLUS", "MINUS", "ASTERISK", "SLASH", "DB_ASTERISK", "DB_SLASH", "PERCENTAGE", "MATMUL",
-        "PLUS_EQUAL", "MINUS_EQUAL", "ASTERISK_EQUAL", "SLASH_EQUAL", "DOT_EQUAL", "COLON_EQUAL", "DB_ASTERISK_EQUAL", "DB_SLASH_EQUAL", "PERCENTAGE_EQUAL", "MATMUL_EQUAL", "SELF_EQUAL",
+        "PLUS", "MINUS", "ASTERISK", "SLASH", "DB_ASTERISK", "DB_SLASH", "PERCENTAGE", "MATMUL", "STRJOIN",
+        "PLUS_EQUAL", "MINUS_EQUAL", "ASTERISK_EQUAL", "SLASH_EQUAL", "DOT_EQUAL", "COLON_EQUAL", "DB_ASTERISK_EQUAL", "DB_SLASH_EQUAL", "PERCENTAGE_EQUAL", "MATMUL_EQUAL", "SELF_EQUAL", "STRJOIN_EQUAL",
         "AMPERSAND", "PIPE", "CARET", "TILDE", "DB_LESS_THAN", "DB_GREATER_THAN",
         "AMPERSAND_EQUAL", "PIPE_EQUAL", "CARET_EQUAL", "TILDE_EQUAL", "DB_LESS_THAN_EQUAL", "DB_GREATER_THAN_EQUAL",
         "LPAREN", "RPAREN", "LBRACK", "RBRACK", "LBRACE", "RBRACE",
         "DOT", "COLON", "SEMICOLON", "COMMA", "EXCLAMATION", "QUESTION",
-        "IF", "ELIF", "ELSE", "WHILE", "FOR", "ASSERT", "USE", "FROM", "AS", "LIKE", "AT", "IN", "TO", "TRY", "CATCH", "DEF", "CLASS", "WITH", "DEL", "IS", "HAS", "RAISE", "BETWEEN", "MATCH", "CASE", "IGNORE",
-        "PASS", "CONTINUE", "BREAK", "GLOBAL",
+        "IF", "ELIF", "ELSE", "WHILE", "FOR", "ASSERT", "USE", "FROM", "AS", "LIKE", "AT", "IN", "TO", "TRY", "CATCH", "DEF", "CLASS", "NAMESPACE", "WITH", "DEL", "IS", "HAS", "RAISE", "BETWEEN", "MATCH", "CASE", "IGNORE", "TIMES", "ANON",
+        "PASS", "CONTINUE", "NEXT", "BREAK", "GLOBAL", "LOCAL",
         "AND", "OR", "NOT",
         "CMP_L", "CMP_G", "CMP_E", "CMP_LE", "CMP_GE", "CMP_NE",
-        "LARROW", "RARROW", "LDARROW", "RDARROW", "LSMARROW", "RSMARROW", "USMARROW", "DSMARROW", "LBARROW", "RBARROW",
+        "LQARROW", "RQARROW", "LARROW", "RARROW", "LDARROW", "RDARROW", "LSMARROW", "RSMARROW", "USMARROW", "DSMARROW", "LBARROW", "RBARROW",
         "DB_ARROW", "DB_DARROW", "DB_SMARROW",
-        "DOLLAR", "HASHTAG", "ELLIPSIS", "OUTPUT",
+        "DOLLAR", "HASHTAG", "ELLIPSIS", "NEW",
         "SUPER_INIT",
-        "NEWLINE", "SPACE"
+        "NEWLINE", "SPACE",
+        "COMMENT"
     }
 
     ignore                  = " \t"
 
-    ignore_comment          = r"=== .*?( ===|\n)"
+    @_(r"=== .*\n")
+    def COMMENT(self, t):
+        t.value = t.value[4:].strip("\n")
+        return t
 
     @_(r'\\\n')
     def LINE_CONTINUATION(self, t):
@@ -71,7 +86,7 @@ class ZinkLexer(Lexer):
         t.value = t.value[2:-1]
         return t
     
-    @_(r'`(?:[^"\\]|\\.)*`')
+    @_(r'`(?:[^"\\]|\\.)*?`')
     def RAWSTRING(self, t):
         t.value = t.value[1:-1]
         return t
@@ -86,6 +101,8 @@ class ZinkLexer(Lexer):
     SELF_EQUAL              = r"@<-"
     SUPER_INIT              = r"@\^"
 
+    LQARROW                 = r"<\?-"
+    RQARROW                 = r"-\?>"
     DB_ARROW                = r"<->"
     DB_DARROW               = r"<=>"
     LDARROW                 = r"<<-"
@@ -113,6 +130,7 @@ class ZinkLexer(Lexer):
     COLON_EQUAL             = r":="
     PERCENTAGE_EQUAL        = r"%="
     MATMUL_EQUAL            = r"@="
+    STRJOIN_EQUAL           = r"\.\.="
 
     DB_ASTERISK             = r"\*\*"
     DB_SLASH                = r"//"
@@ -122,6 +140,7 @@ class ZinkLexer(Lexer):
     SLASH                   = r"/"
     PERCENTAGE              = r"%"
     MATMUL                  = r"@"
+    STRJOIN                 = r"\.\."
 
     AMPERSAND_EQUAL         = r"&="
     PIPE_EQUAL              = r"\|="
@@ -179,7 +198,9 @@ class ZinkLexer(Lexer):
     ID["catch"]             = "CATCH"
     ID["pass"]              = "PASS"
     ID["continue"]          = "CONTINUE"
+    ID["next"]              = "NEXT"
     ID["global"]            = "GLOBAL"
+    ID["local"]             = "LOCAL"
     ID["break"]             = "BREAK"
     ID["True"]              = "TRUE"
     ID["False"]             = "FALSE"
@@ -198,6 +219,9 @@ class ZinkLexer(Lexer):
     ID["match"]             = "MATCH"
     ID["case"]              = "CASE"
     ID["ignore"]            = "IGNORE"
+    ID["times"]             = "TIMES"
+    ID["namespace"]         = "NAMESPACE"
+    ID["anon"]              = "ANON"
 
     @_(r"0x[0-9a-fA-F_]+", r"0b[01_]+", r"[0-9_]+", r"[0-9_]\.[0-9_]", r"\.[0-9_]")
     def NUMBER(self, t):
@@ -211,9 +235,9 @@ class ZinkLexer(Lexer):
             t.value = int(t.value)
         return t
     
-    @_(r"\n+")
+    @_(r"\n")
     def NEWLINE(self, t):
-        self.lineno += len(t.value)
+        self.lineno += 1
         return t
     
     def find_column(text, token):
@@ -227,35 +251,44 @@ class ZinkParser(Parser):
     debug = False
     debugfile = "parser.txt" if debug else None
     log = FilteredLogger(sys.stderr, exit=not debug)
+    had_errors = False
+
+    def __init__(self, ignore_obsolete: bool = False, include_comments: bool = False, include_empty_lines: bool = False):
+        super().__init__()
+        self.ignore_obsolete = ignore_obsolete
+        self.include_comments = include_comments
+        self.include_empty_lines = include_empty_lines
 
     def error(self, token):
-        sys.stderr.write("\b\b\b\b[!] ")
+        self.had_errors = True
         if token:
             lineno = getattr(token, "lineno", 0)
             if lineno:
-                sys.stderr.write(f"Error at line {lineno} (token \"{token.type}\")\n")
+                sys.stderr.write(_error(f"Token \"{token.type}\" at line {lineno}\n"))
             else:
-                sys.stderr.write(f"Error (token \"{token.type}\")\n")
+                sys.stderr.write(_error(f"Token \"{token.type}\"\n"))
         else:
-            sys.stderr.write("Unexpected EOF\n")
-        exit(1)
+            sys.stderr.write(_error("Unexpected end of file\n"))
+    
+    def reset_errors(self):
+        self.had_errors = False
 
     tokens = ZinkLexer.tokens
 
     precedence = (
         ("right", "EQUAL"),
         ("right", "GENERATOR", "TERNARY"),
-        ("nonassoc", "PLUS_EQUAL", "MINUS_EQUAL", "ASTERISK_EQUAL", "SLASH_EQUAL", "DOT_EQUAL", "PERCENTAGE_EQUAL", "AMPERSAND_EQUAL", "MATMUL_EQUAL", "PIPE_EQUAL", "CARET_EQUAL", "TILDE_EQUAL", "DB_LESS_THAN_EQUAL", "DB_GREATER_THAN_EQUAL"),
+        ("nonassoc", "PLUS_EQUAL", "MINUS_EQUAL", "ASTERISK_EQUAL", "SLASH_EQUAL", "DOT_EQUAL", "PERCENTAGE_EQUAL", "AMPERSAND_EQUAL", "MATMUL_EQUAL", "PIPE_EQUAL", "CARET_EQUAL", "TILDE_EQUAL", "DB_LESS_THAN_EQUAL", "DB_GREATER_THAN_EQUAL", "STRJOIN_EQUAL"),
         ("right", "NOT"),
         ("left", "AND", "OR"),
         ("nonassoc", "CMP_L", "CMP_G", "CMP_E", "CMP_LE", "CMP_GE", "CMP_NE", "SAME", "IN"),
         ("left", "INDEX"),
-        ("left", "PLUS", "MINUS"),
+        ("left", "PLUS", "MINUS", "STRJOIN"),
         ("left", "ASTERISK", "SLASH", "DB_ASTERISK", "DB_SLASH", "PERCENTAGE", "MATMUL", "AMPERSAND", "PIPE", "CARET", "DB_LESS_THAN", "DB_GREATER_THAN", "COLON_EQUAL"),
         ("right", "UNARY_PLUS", "UNARY_MINUS", "STRING_UPPER", "STRING_LOWER", "LENGTH", "TYPE", "TILDE"),
         ("left", "INCREMENT", "DECREMENT"),
         ("left", "AS", "LIKE"),
-        ("left", "MEMBER", "DOT", "LPAREN", "EXCLAMATION"),
+        ("left", "MEMBER", "DOT", "LPAREN", "EXCLAMATION")
     )
 
     @_("stmts")
@@ -355,15 +388,18 @@ class ZinkParser(Parser):
     def farg(self, p):
         return ("default_typed_arg", p.ID, p.type, p.expr)
     
+    @_("SLASH")
+    def farg(self, p):
+        return ("slash_arg",)
+    
+    @_("ASTERISK")
+    def farg(self, p):
+        return ("asterisk_arg",)
+    
     @_("fargs COMMA farg",
        "fargs COMMA MATMUL farg",
-       "fargs COMMA CARET farg")
-    def fargs(self, p):
-        if hasattr(p, "MATMUL"): return p.fargs + [("func_assign_self", p.farg)]
-        elif hasattr(p, "CARET"): return p.fargs + [("func_assign_super", p.farg)]
-        return p.fargs + [p.farg]
-    
-    @_("fargs COMMA NEWLINE farg",
+       "fargs COMMA CARET farg",
+       "fargs COMMA NEWLINE farg",
        "fargs COMMA NEWLINE MATMUL farg",
        "fargs COMMA NEWLINE CARET farg")
     def fargs(self, p):
@@ -415,6 +451,22 @@ class ZinkParser(Parser):
     def kwargs(self, p):
         return [p.kwarg]
     
+    @_("ID")
+    def dotid(self, p):
+        return ("dotid", p.ID)
+    
+    @_("RAWSTRING")
+    def dotid(self, p):
+        return ("dotid", p.RAWSTRING)
+
+    @_("ID DOT dotid")
+    def dotid(self, p):
+        return ("dotid_dot", p.ID, p.dotid)
+
+    @_("RAWSTRING DOT dotid")
+    def dotid(self, p):
+        return ("dotid_dot", p.RAWSTRING, p.dotid)
+    
     @_("ELIF expr end program DOT",
        "ELIF expr end program DOT end")
     def if_elif(self, p):
@@ -451,12 +503,16 @@ class ZinkParser(Parser):
     
     @_("end")
     def stmt(self, p):
-        return None
+        return ("EMPTY_LINE",) if self.include_empty_lines else None
     
     @_("expr end")
     def stmt(self, p):
         return p.expr
     
+    @_("COMMENT")
+    def stmt(self, p):
+        return ("COMMENT", p.COMMENT) if self.include_comments else None
+    
     @_("targs EQUAL args end")
     def stmt(self, p):
         return ("set", p.targs, p.args)
@@ -525,6 +581,10 @@ class ZinkParser(Parser):
     def stmt(self, p):
         return ("set_self", p.expr)
     
+    @_("expr STRJOIN_EQUAL expr end")
+    def stmt(self, p):
+        return ("set_strjoin", p.expr0, p.expr1)
+    
     @_("expr TO expr end")
     def stmt(self, p):
         return ("set_cast", p.expr0, p.expr1)
@@ -545,21 +605,21 @@ class ZinkParser(Parser):
     def stmt(self, p):
         return ("raise", p.expr)
     
-    @_("USE ID end")
+    @_("USE dotid end")
     def stmt(self, p):
-        return ("use", p.ID)
+        return ("use", p.dotid)
     
-    @_("USE ID AS ID end")
+    @_("USE dotid AS dotid end")
     def stmt(self, p):
-        return ("use_as", p.ID0, p.ID1)
+        return ("use_as", p.dotid0, p.dotid1)
     
-    @_("USE ID FROM ID end")
+    @_("USE dotid FROM dotid end")
     def stmt(self, p):
-        return ("use_from", p.ID0, p.ID1)
+        return ("use_from", p.dotid0, p.dotid1)
     
-    @_("USE ID AS ID FROM ID end")
+    @_("USE dotid AS dotid FROM dotid end")
     def stmt(self, p):
-        return ("use_as_from", p.ID0, p.ID1, p.ID2)
+        return ("use_as_from", p.dotid0, p.dotid1, p.dotid2)
     
     @_("WHILE expr end program DOT end")
     def stmt(self, p):
@@ -573,6 +633,26 @@ class ZinkParser(Parser):
     def stmt(self, p):
         return ("for_at", p.args, p.expr0, p.expr1, p.program)
     
+    @_("FOR args RARROW expr end program DOT ELSE end program DOT end")
+    def stmt(self, p):
+        return ("for_else", p.args, p.expr, p.program0, p.program1)
+    
+    @_("FOR args AT expr RARROW expr end program DOT ELSE end program DOT end")
+    def stmt(self, p):
+        return ("for_at_else", p.args, p.expr0, p.expr1, p.program0, p.program1)
+    
+    @_("FOR var TO expr end program DOT end")
+    def stmt(self, p):
+        return ("for_to", p.var, p.expr, p.program)
+    
+    @_("FOR var FROM expr TO expr end program DOT end")
+    def stmt(self, p):
+        return ("for_from_to", p.var, p.expr0, p.expr1, p.program)
+    
+    @_("TIMES expr end program DOT end")
+    def stmt(self, p):
+        return ("times", p.expr, p.program)
+    
     @_("IF expr end program DOT end")
     def stmt(self, p):
         return ("if", p.expr, p.program)
@@ -626,10 +706,10 @@ class ZinkParser(Parser):
     @_("PASS end")
     def stmt(self, p):
         return ("pass",)
-    
-    @_("CONTINUE end")
+
+    @_("NEXT end")
     def stmt(self, p):
-        return ("continue",)
+        return ("next",)
     
     @_("BREAK end")
     def stmt(self, p):
@@ -639,21 +719,91 @@ class ZinkParser(Parser):
     def stmt(self, p):
         return ("global", p.var)
     
-    @_("DEF ID end program DOT")
+    @_("EXCLAMATION LOCAL var end")
     def stmt(self, p):
-        return (f"func_def_untyped", p.ID, [], p.program)
+        return ("nonlocal", p.var)
     
-    @_("DEF ID LPAREN fargs RPAREN end program DOT")
+    @_("LOCAL var end")
     def stmt(self, p):
-        return (f"func_def_untyped", p.ID, p.fargs, p.program)
+        return ("local", p.var)
     
-    @_("DEF ID COLON type end program DOT")
+    @_("GLOBAL targs EQUAL args end")
     def stmt(self, p):
-        return (f"func_def", p.ID, [], p.type, p.program)
+        return ("global_set", p.targs, p.args)
     
-    @_("DEF ID LPAREN fargs RPAREN COLON type end program DOT")
+    @_("EXCLAMATION LOCAL targs EQUAL args end")
     def stmt(self, p):
-        return (f"func_def", p.ID, p.fargs, p.type, p.program)
+        return ("nonlocal_set", p.targs, p.args)
+    
+    @_("LOCAL targs EQUAL args end")
+    def stmt(self, p):
+        return ("local_set", p.targs, p.args)
+    
+    @_("DEF dotid end program DOT",
+       "DEF MATMUL COLON dotid end program DOT",
+       "DEF QUESTION COLON dotid end program DOT",
+       "DEF QUESTION MATMUL COLON dotid end program DOT",
+       "DEF dotid LPAREN fargs RPAREN end program DOT",
+       "DEF MATMUL COLON dotid LPAREN fargs RPAREN end program DOT",
+       "DEF QUESTION COLON dotid LPAREN fargs RPAREN end program DOT",
+       "DEF QUESTION MATMUL COLON dotid LPAREN fargs RPAREN end program DOT",
+       "DEF dotid COLON type end program DOT",
+       "DEF MATMUL COLON dotid COLON type end program DOT",
+       "DEF QUESTION COLON dotid COLON type end program DOT",
+       "DEF QUESTION MATMUL COLON dotid COLON type end program DOT",
+       "DEF dotid LPAREN fargs RPAREN COLON type end program DOT",
+       "DEF MATMUL COLON dotid LPAREN fargs RPAREN COLON type end program DOT",
+       "DEF QUESTION COLON dotid LPAREN fargs RPAREN COLON type end program DOT",
+       "DEF QUESTION MATMUL COLON dotid LPAREN fargs RPAREN COLON type end program DOT",
+       "LOCAL DEF dotid end program DOT",
+       "LOCAL DEF MATMUL COLON dotid end program DOT",
+       "LOCAL DEF QUESTION COLON dotid end program DOT",
+       "LOCAL DEF QUESTION MATMUL COLON dotid end program DOT",
+       "LOCAL DEF dotid LPAREN fargs RPAREN end program DOT",
+       "LOCAL DEF MATMUL COLON dotid LPAREN fargs RPAREN end program DOT",
+       "LOCAL DEF QUESTION COLON dotid LPAREN fargs RPAREN end program DOT",
+       "LOCAL DEF QUESTION MATMUL COLON dotid LPAREN fargs RPAREN end program DOT",
+       "LOCAL DEF dotid COLON type end program DOT",
+       "LOCAL DEF MATMUL COLON dotid COLON type end program DOT",
+       "LOCAL DEF QUESTION COLON dotid COLON type end program DOT",
+       "LOCAL DEF QUESTION MATMUL COLON dotid COLON type end program DOT",
+       "LOCAL DEF dotid LPAREN fargs RPAREN COLON type end program DOT",
+       "LOCAL DEF MATMUL COLON dotid LPAREN fargs RPAREN COLON type end program DOT",
+       "LOCAL DEF QUESTION COLON dotid LPAREN fargs RPAREN COLON type end program DOT",
+       "LOCAL DEF QUESTION MATMUL COLON dotid LPAREN fargs RPAREN COLON type end program DOT")
+    def stmt(self, p):
+        if hasattr(p, "type"):
+            return (f"func_def{"_local" if hasattr(p, "LOCAL") else ""}{"_async" if hasattr(p, "QUESTION") else ""}{"_self" if hasattr(p, "MATMUL") else ""}", p.dotid, getattr(p, "fargs", []), p.type, p.program)
+        else:
+            return (f"func_def{"_local" if hasattr(p, "LOCAL") else ""}{"_async" if hasattr(p, "QUESTION") else ""}{"_self" if hasattr(p, "MATMUL") else ""}_untyped", p.dotid, getattr(p, "fargs", []), p.program)
+    
+    @_("SLASH ID end program DOT",
+       "SLASH ID fargs end program DOT",
+       "SLASH QUESTION ID end program DOT",
+       "SLASH QUESTION ID fargs end program DOT",
+       "SLASH ASTERISK end program DOT",
+       "SLASH ASTERISK fargs end program DOT",
+       "SLASH QUESTION ASTERISK end program DOT",
+       "SLASH QUESTION ASTERISK fargs end program DOT",
+       "SLASH PLUS end program DOT",
+       "SLASH PLUS fargs end program DOT",
+       "SLASH QUESTION PLUS end program DOT",
+       "SLASH QUESTION PLUS fargs end program DOT",
+       "SLASH MINUS end program DOT",
+       "SLASH MINUS fargs end program DOT",
+       "SLASH QUESTION MINUS end program DOT",
+       "SLASH QUESTION MINUS fargs end program DOT",
+       "SLASH EXCLAMATION end program DOT",
+       "SLASH EXCLAMATION fargs end program DOT",
+       "SLASH QUESTION EXCLAMATION end program DOT",
+       "SLASH QUESTION EXCLAMATION fargs end program DOT")
+    def stmt(self, p):
+        if hasattr(p, "ID"): func_type = p.ID
+        elif hasattr(p, "ASTERISK"): func_type = "init"
+        elif hasattr(p, "PLUS"): func_type = "enter"
+        elif hasattr(p, "MINUS"): func_type = "exit"
+        elif hasattr(p, "EXCLAMATION"): func_type = "call"
+        return (f"func_def{"_async" if hasattr(p, "QUESTION") else ""}__", func_type, getattr(p, "fargs", []), p.program)
     
     @_("CLASS ID end program DOT")
     def stmt(self, p):
@@ -663,17 +813,20 @@ class ZinkParser(Parser):
     def stmt(self, p):
         return ("class_def_from", p.ID0, p.ID1, p.program)
     
+    @_("NAMESPACE ID end program DOT")
+    def stmt(self, p):
+        return ("namespace_def", p.ID, p.program)
+    
     @_("WITH expr AS expr end program DOT")
     def stmt(self, p):
         return ("with", p.expr0, p.expr1, p.program)
     
-    @_("LARROW end")
+    @_("LARROW end",
+       "LARROW expr end",
+       "LQARROW end",
+       "LQARROW expr end")
     def stmt(self, p):
-        return ("return", None)
-    
-    @_("LARROW expr end")
-    def stmt(self, p):
-        return ("return", p.expr)
+        return ("yield" if hasattr(p, "LQARROW") else "return", getattr(p, "expr", ("NONE",)))
     
     @_("DEL expr end")
     def stmt(self, p):
@@ -715,93 +868,72 @@ class ZinkParser(Parser):
     def expr(self, p):
         return ("NONE",)
     
-    @_("expr LPAREN fcargs RPAREN")
+    @_("expr LPAREN fcargs RPAREN",
+       "expr LPAREN NEWLINE fcargs NEWLINE RPAREN",
+       "expr EXCLAMATION")
     def func(self, p):
-        return ("func", p.expr, p.fcargs)
+        return ("func", p.expr, getattr(p, "fcargs", []))
     
-    @_("expr LPAREN NEWLINE fcargs NEWLINE RPAREN")
+    @_("expr DOT var DOLLAR LPAREN fcargs RPAREN",
+       "expr DOT var DOLLAR LPAREN NEWLINE fcargs NEWLINE RPAREN",
+       "expr DOT var DOLLAR EXCLAMATION")
     def func(self, p):
-        return ("func", p.expr, p.fcargs)
-    
-    @_("expr EXCLAMATION")
-    def func(self, p):
-        return ("func", p.expr, [])
-    
+        return ("func_self", p.expr, p.var, getattr(p, "fcargs", []))
+
     @_("func")
     def expr(self, p):
         return p.func
     
-    @_("LBRACK args RBRACK")
+    @_("LBRACK RBRACK",
+       "LBRACK args RBRACK",
+       "LBRACK NEWLINE args NEWLINE RBRACK")
     def tuple(self, p):
-        return ("tuple", p.args)
-    
-    @_("LBRACK NEWLINE args NEWLINE RBRACK")
-    def tuple(self, p):
-        return ("tuple", p.args)
-    
-    @_("LBRACK RBRACK")
-    def tuple(self, p):
-        return ("tuple", [])
+        return ("tuple", getattr(p, "args", []))
     
     @_("tuple")
     def expr(self, p):
         return p.tuple
     
-    @_("LBRACK args COMMA RBRACK")
+    @_("LBRACK COMMA RBRACK",
+       "LBRACK args COMMA RBRACK",
+       "LBRACK NEWLINE args COMMA NEWLINE RBRACK")
     def list(self, p):
-        return ("list", p.args)
-    
-    @_("LBRACK NEWLINE args COMMA NEWLINE RBRACK")
-    def list(self, p):
-        return ("list", p.args)
-    
-    @_("LBRACK COMMA RBRACK")
-    def list(self, p):
-        return ("list", [])
+        return ("list", getattr(p, "args", []))
     
     @_("list")
     def expr(self, p):
         return p.list
     
-    @_("LBRACK kwargs RBRACK")
+    @_("LBRACK EQUAL RBRACK",
+       "LBRACK kwargs RBRACK",
+       "LBRACK NEWLINE kwargs NEWLINE RBRACK")
     def dict(self, p):
-        return ("dict", p.kwargs)
-    
-    @_("LBRACK NEWLINE kwargs NEWLINE RBRACK")
-    def dict(self, p):
-        return ("dict", p.kwargs)
-    
-    @_("LBRACK EQUAL RBRACK")
-    def dict(self, p):
-        return ("dict", [])
+        return ("dict", getattr(p, "kwargs", []))
     
     @_("dict")
     def expr(self, p):
         return p.dict
     
-    @_("LBRACK expr DB_ARROW expr RBRACK")
+    @_("LBRACK expr DB_ARROW expr RBRACK",
+       "LBRACK expr DB_ARROW expr RPAREN",
+       "LPAREN expr DB_ARROW expr RBRACK",
+       "LPAREN expr DB_ARROW expr RPAREN",
+       "LBRACK expr DB_ARROW expr COMMA expr RBRACK",
+       "LBRACK expr DB_ARROW expr COMMA expr RPAREN",
+       "LPAREN expr DB_ARROW expr COMMA expr RBRACK",
+       "LPAREN expr DB_ARROW expr COMMA expr RPAREN")
     def range(self, p):
-        return ("range_inc_inc", p.expr0, p.expr1, ("NUMBER", "1"))
+        return (f"range{"_inc" if hasattr(p, "LBRACK") else "_exc"}{"_inc" if hasattr(p, "RBRACK") else "_exc"}", p.expr0, p.expr1, getattr(p, "expr2", ("NUMBER", "1")))
     
-    @_("LBRACK expr DB_ARROW expr RPAREN")
+    @_("LBRACE expr DB_ARROW expr RBRACE",
+       "LBRACE expr DB_ARROW expr COMMA expr RBRACE")
     def range(self, p):
-        return ("range_inc_exc", p.expr0, p.expr1, ("NUMBER", "1"))
+        return ("range", p.expr0, p.expr1, getattr(p, "expr2", ("NUMBER", "1")))
     
-    @_("LPAREN expr DB_ARROW expr RBRACK")
+    @_("LBRACE RARROW expr RBRACE",
+       "LBRACE RARROW expr COMMA expr RBRACE")
     def range(self, p):
-        return ("range_exc_inc", p.expr0, p.expr1, ("NUMBER", "1"))
-    
-    @_("LPAREN expr DB_ARROW expr RPAREN")
-    def range(self, p):
-        return ("range_exc_exc", p.expr0, p.expr1, ("NUMBER", "1"))
-    
-    @_("LBRACE expr DB_ARROW expr RBRACE")
-    def range(self, p):
-        return ("range", p.expr0, p.expr1, ("NUMBER", "1"))
-    
-    @_("LBRACE RARROW expr RBRACE")
-    def range(self, p):
-        return ("range", ("NUMBER", "0"), p.expr, ("NUMBER", "1"))
+        return ("range", ("NUMBER", "0"), p.expr0 if hasattr(p, "expr0") else p.expr, getattr(p, "expr1", ("NUMBER", "1")))
     
     @_("range")
     def expr(self, p):
@@ -863,6 +995,10 @@ class ZinkParser(Parser):
     def expr(self, p):
         return ("modulo", p.expr0, p.expr1)
     
+    @_("expr STRJOIN expr")
+    def expr(self, p):
+        return ("strjoin", p.expr0, p.expr1)
+    
     @_("expr DB_PLUS %prec INCREMENT")
     def expr(self, p):
         return ("inc_after", p.expr)
@@ -949,7 +1085,7 @@ class ZinkParser(Parser):
 
     @_("LPAREN expr RPAREN")
     def expr(self, p):
-        return (p.expr)
+        return ("PAREN", p.expr)
     
     @_("expr LBRACK expr RBRACK %prec INDEX")
     def expr(self, p):
@@ -1043,6 +1179,28 @@ class ZinkParser(Parser):
     def expr(self, p):
         return ("between", p.expr0, p.expr1, p.expr2)
     
+    @_("ANON end program DOT",
+       "ANON MATMUL end program DOT",
+       "ANON QUESTION end program DOT",
+       "ANON QUESTION MATMUL end program DOT",
+       "ANON LPAREN fargs RPAREN end program DOT",
+       "ANON MATMUL LPAREN fargs RPAREN end program DOT",
+       "ANON QUESTION LPAREN fargs RPAREN end program DOT",
+       "ANON QUESTION MATMUL LPAREN fargs RPAREN end program DOT",
+       "ANON COLON type end program DOT",
+       "ANON MATMUL COLON type end program DOT",
+       "ANON QUESTION COLON type end program DOT",
+       "ANON QUESTION MATMUL COLON type end program DOT",
+       "ANON LPAREN fargs RPAREN COLON type end program DOT",
+       "ANON MATMUL LPAREN fargs RPAREN COLON type end program DOT",
+       "ANON QUESTION LPAREN fargs RPAREN COLON type end program DOT",
+       "ANON QUESTION MATMUL LPAREN fargs RPAREN COLON type end program DOT")
+    def expr(self, p):
+        if hasattr(p, "type"):
+            return (f"func_def_anon{"_async" if hasattr(p, "QUESTION") else ""}{"_self" if hasattr(p, "MATMUL") else ""}", getattr(p, "fargs", []), p.type, p.program)
+        else:
+            return (f"func_def_anon{"_async" if hasattr(p, "QUESTION") else ""}{"_self" if hasattr(p, "MATMUL") else ""}_untyped", getattr(p, "fargs", []), p.program)
+    
     @_("MATMUL")
     def expr(self, p):
         return ("self",)
